@@ -1,11 +1,11 @@
 /* ber-decoder.c - Basic Encoding Rules Decoder
- *      Copyright (C) 2001, 2004 g10 Code GmbH
+ *      Copyright (C) 2001, 2004, 2006 g10 Code GmbH
  *
  * This file is part of KSBA.
  *
  * KSBA is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * KSBA is distributed in the hope that it will be useful,
@@ -14,8 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -767,7 +766,7 @@ static gpg_error_t
 decoder_next (BerDecoder d)
 {
   struct tag_info ti;
-  AsnNode node = NULL;
+  AsnNode node;
   gpg_error_t err;
   DECODER_STATE ds = d->ds;
   int debug = d->debug;
@@ -812,8 +811,8 @@ decoder_next (BerDecoder d)
     {
       if (!d->image.buf)
         {
-          /* we need some extra bytes to store the stuff we read ahead
-             at the end of the module which is later pushed back */
+          /* We need some extra bytes to store the stuff we read ahead
+             at the end of the module which is later pushed back. */
           d->image.length = ti.length + 100;
           d->image.used = 0;
           d->image.buf = xtrymalloc (d->image.length);
@@ -823,6 +822,7 @@ decoder_next (BerDecoder d)
 
       if (ti.nhdr + d->image.used >= d->image.length)
         return set_error (d, NULL, "image buffer too short to store the tag");
+
       memcpy (d->image.buf + d->image.used, ti.buf, ti.nhdr);
       d->image.used += ti.nhdr;
     }
@@ -1138,9 +1138,12 @@ _ksba_ber_decoder_decode (BerDecoder d, const char *start_name,
       int n, c;
 
       node = d->val.node;
-      if (node && d->use_image)
+      /* Fixme: USE_IMAGE is only not used with the ber-dump utility
+         and thus of no big use.  We should remove the other code
+         paths and dump ber-dump.c. */
+      if (d->use_image)
         {
-          if (!d->val.is_endtag)
+          if (node && !d->val.is_endtag)
             { /* We don't have nodes for the end tag - so don't store it */
               node->off = (ksba_reader_tell (d->reader)
                            - d->val.nhdr - startoff);
@@ -1197,11 +1200,14 @@ _ksba_ber_decoder_decode (BerDecoder d, const char *start_name,
   if (gpg_err_code (err) == GPG_ERR_EOF)
     err = 0;
 
+  if (err)
+    xfree (d->image.buf);
+
   if (r_root && !err)
     {
       if (!d->image.buf)
         { /* Not even the first node available - return eof */
-          /* Fixme: release d->root */
+	  _ksba_asn_release_nodes (d->root);
           d->root = NULL;
           err = gpg_error (GPG_ERR_EOF);
         }
