@@ -5,7 +5,7 @@
  *
  * KSBA is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * KSBA is distributed in the hope that it will be useful,
@@ -14,9 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
- * USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 /*
@@ -753,11 +751,15 @@ _ksba_cms_parse_signed_data_part_2 (ksba_cms_t cms)
          signature.  Thus we have to allow for EOF here */
       if (gpg_err_code (err) == GPG_ERR_EOF)
         {
+	  xfree (si);
           err = 0;
           break;
         }
       if (err)
-        return err;
+	{
+	  xfree (si);
+	  return err;
+	}
 
       *si_tail = si;
       si_tail = &si->next;
