@@ -5,7 +5,7 @@
  *
  * KSBA is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Fountion; either version 2 of the License, or
+ * the Free Software Fountion; either version 3 of the License, or
  * (at your option) any later version.
  *
  * KSBA is distributed in the hope that it will be useful,
@@ -14,9 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
- * USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -59,7 +57,7 @@ do_hash (ksba_crl_t crl, const void *buffer, size_t length)
                            crl->hashbuf.buffer, crl->hashbuf.used);
           crl->hashbuf.used = 0;
         }
-      buffer += n;
+      buffer = (const char *)buffer + n;
       length -= n;
     }
 }
