@@ -5,7 +5,7 @@
  *
  * KSBA is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Fountion; either version 2 of the License, or
+ * the Free Software Fountion; either version 3 of the License, or
  * (at your option) any later version.
  *
  * KSBA is distributed in the hope that it will be useful,
@@ -14,9 +14,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
- * USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -1200,8 +1198,8 @@ parse_single_response (ksba_ocsp_t ocsp,
         ; /* Cope with zero length objects. */
       else if (*datalen && !**data)
         { /* Skip the NULL. */
-          *datalen--;
-          *data++;
+          (*datalen)--;
+          (*data)++;
         }
       else
         return gpg_error (GPG_ERR_INV_OBJ);
@@ -1261,8 +1259,8 @@ parse_single_response (ksba_ocsp_t ocsp,
         ; /* Cope with zero length objects. */
       else if (!**data)
         { /* Skip the NULL. */
-          *datalen--;
-          *data++;
+          (*datalen)--;
+          (*data)++;
         }
       else /* The comment indicates that an enumeration may come here. */ 
         {
